// ***********************************************************************
//
//            Vite: A C++ library for distributed-memory graph clustering 
//                  using MPI+OpenMP
// 
//               Daniel Chavarria (daniel.chavarria@pnnl.gov)
//               Antonino Tumeo (antonino.tumeo@pnnl.gov)
//               Mahantesh Halappanavar (hala@pnnl.gov)
//               Pacific Northwest National Laboratory	
//
//               Hao Lu (luhowardmark@wsu.edu)
//               Sayan Ghosh (sayan.ghosh@wsu.edu)
//               Ananth Kalyanaraman (ananth@eecs.wsu.edu)
//               Washington State University
//
// ***********************************************************************
//
//       Copyright (2017) Battelle Memorial Institute
//                      All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// ************************************************************************
#include <cmath>

#include <fstream>
#include <iostream>
#include <sstream>
#include <vector>

#include "dimacs.hpp"
#include "../utils.hpp"

void loadDimacsFile(Graph *&g, const std::string &fileName)
{
  std::ifstream ifs;

  ifs.open(fileName.c_str());
  if (!ifs) {
    std::cerr << "Error opening Dimacs format file: " << fileName << std::endl;
    exit(EXIT_FAILURE);
  }

  std::string line;
  char comment;

  do {
    std::getline(ifs, line);
    comment = line[0];
  } while (comment == 'c');

  long int numVertices, numEdges, value = 0L;
  {
    std::string p, sp;

    std::istringstream iss(line);

    iss >> p >> sp >> numVertices >> numEdges;
    if (!iss || iss.fail() || p[0] != 'p') {
      std::cerr << "Error parsing DIMACS format: " << line << std::endl;
      std::cerr << "p: " << p << ", p[0]: " << p[0] << ", sp: " << sp <<
	", numVertices: " << numVertices << ", numEdges: " << numEdges << std::endl;
      exit(EXIT_FAILURE);
    }
  }

  std::cout << "Loading DIMACS file: " << fileName << ", numvertices: " << numVertices <<
    ", numEdges: " << numEdges << std::endl;

  std::vector<GraphElem> edgeCount(numVertices + 1L);
  std::vector<GraphElemPair> edgeList(numEdges * 2L);

  for (long int i = 0; i < numEdges; i++) {
    std::string label;
    long int source, dest;
    double weight;

    ifs >> label >> source >> dest >> weight;

    if (!ifs || ifs.fail()) {
      std::cerr << "Error parsing DIMACS edge" << std::endl;
      exit(EXIT_FAILURE);
    }

    if ((i % 100000L) == 0L)
      std::cout << "Reading edge: (" << source << ", " << dest << ") with weight: " <<
	weight << std::endl;

    GraphElemPair &fedge = edgeList[2L * i];
    GraphElemPair &bedge = edgeList[(2L * i) + 1L];

    fedge.first = source - 1L;
    fedge.second = dest - 1L;

    bedge.first = dest - 1L;
    bedge.second = source - 1L;

    edgeCount[fedge.first + 1L]++;
    edgeCount[bedge.first + 1L]++;
  }

  ifs.close();

  numEdges *= 2L;
  g = new Graph(numVertices, numEdges);

  processGraphData(*g, edgeCount, edgeList, numVertices, numEdges);
} // loadDimacsFile
