// ***********************************************************************
//
//            Vite: A C++ library for distributed-memory graph clustering 
//                  using MPI+OpenMP
// 
//               Daniel Chavarria (daniel.chavarria@pnnl.gov)
//               Antonino Tumeo (antonino.tumeo@pnnl.gov)
//               Mahantesh Halappanavar (hala@pnnl.gov)
//               Pacific Northwest National Laboratory	
//
//               Hao Lu (luhowardmark@wsu.edu)
//               Sayan Ghosh (sayan.ghosh@wsu.edu)
//               Ananth Kalyanaraman (ananth@eecs.wsu.edu)
//               Washington State University
//
// ***********************************************************************
//
//       Copyright (2017) Battelle Memorial Institute
//                      All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// ************************************************************************

#include <fstream>
#include <iostream>
#include <numeric>
#include <sstream>
#include <vector>
#include <utility>

#include "simple.hpp"
#include "../utils.hpp"

void loadSimpleFileUn(Graph *&g, const std::string &fileName)
{
  std::ifstream ifs;

  double t0, t1, t2, t3;

  t0 = mytimer();

  ifs.open(fileName.c_str(), std::ifstream::in);
  if (!ifs) {
    std::cerr << "Error opening Simple format file: " << fileName << std::endl;
    exit(EXIT_FAILURE);
  }

  std::string line;
  long int maxVertex = -1L, numEdges = 0L, numVertices;

  do {
    long int v0, v1;

    std::getline(ifs, line);
		if(line[0]=='#' || line[0]=='%'){
			std::cout<< "Skip Line" <<std::endl;
			continue;
		}
	  //std::cout << "Last line read: " << line << std::endl;

    std::istringstream iss(line);

    iss >> v0 >> v1;
	  //std::cout << v0 << ","  << v1 << std::endl;

    //assert(iss.eof());

    if (v0 > maxVertex)
      maxVertex = v0;

    if (v1 > maxVertex)
      maxVertex = v1;

    numEdges++;
  } while (!ifs.eof());

  std::cout << "Last line read: " << line << std::endl;
  numEdges--;  // Do not consider last line
  numEdges *=2;

  numVertices = maxVertex + 1L;

  t1 = mytimer();

  std::cout << "Loading Simple file: " << fileName << ", numvertices: " << numVertices <<
    ", numEdges: " << numEdges << std::endl;
  std::cout << "Edge & vertex count time: " << (t1 - t0) << std::endl;

  t2 = mytimer();

  ifs.close();
  ifs.open(fileName.c_str(), std::ifstream::in);

  std::vector<GraphElem> edgeCount(numVertices + 1L);
  std::vector<GraphElemPair> edgeList(numEdges);

  for (long int i = 0L; i < numEdges; i+=2) {
    long int v0, v1;

    std::getline(ifs, line);
		if(line[0]=='#' || line[0]=='%'){
			std::cout<< "Skip Line" <<std::endl;
			i--;
			continue;
		}
    std::istringstream iss(line);

    iss >> v0 >> v1;
//    assert(iss.eof());

    edgeList[i].first = v0;
    edgeList[i].second = v1;
    edgeList[i+1].first = v1;
    edgeList[i+1].second = v0;

    edgeCount[v0 + 1L]++;
    edgeCount[v1+1L]++;
  }

  ifs.close();

  t3 = mytimer();

  std::cout << "Edge read time: " << (t3 - t2) << std::endl;

  t2 = mytimer();

  std::cout << "Before allocating graph" << std::endl;

  g = new Graph(numVertices, numEdges);

  assert(g);

  std::cout << "Allocated graph: " << g << std::endl;

  processGraphData(*g, edgeCount, edgeList, numVertices, numEdges);

  t3 = mytimer();  

  std::cout << "Graph populate time: " << (t3 - t2) << std::endl;
  std::cout << "Total I/O time: " << (t3 - t0) << std::endl;
} // loadSimpleFile
