// ***********************************************************************
//
//            Vite: A C++ library for distributed-memory graph clustering 
//                  using MPI+OpenMP
// 
//               Daniel Chavarria (daniel.chavarria@pnnl.gov)
//               Antonino Tumeo (antonino.tumeo@pnnl.gov)
//               Mahantesh Halappanavar (hala@pnnl.gov)
//               Pacific Northwest National Laboratory	
//
//               Hao Lu (luhowardmark@wsu.edu)
//               Sayan Ghosh (sayan.ghosh@wsu.edu)
//               Ananth Kalyanaraman (ananth@eecs.wsu.edu)
//               Washington State University
//
// ***********************************************************************
//
//       Copyright (2017) Battelle Memorial Institute
//                      All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// ************************************************************************

#ifndef __DISTGRAPH_H
#define __DISTGRAPH_H
///
#include <cassert>

#include <algorithm>

#include "graph.hpp"

typedef std::vector<GraphElem> PartRanges;

class DistGraph {
protected:
  GraphElem totalNumVertices;
  GraphElem totalNumEdges;
  GraphElem blkSize;
  Graph *localGraph;
//  PartRanges *parts;

public:
  DistGraph(const GraphElem tnv, const GraphElem tne);
  DistGraph(const DistGraph &othis);
  ~DistGraph();

  GraphElem getTotalNumVertices() const;
  GraphElem getTotalNumEdges() const;

  void createLocalGraph(const GraphElem lnv, const GraphElem lne,
			const PartRanges *oparts = NULL);
  Graph &getLocalGraph();
  const Graph &getLocalGraph() const;
  GraphElem &getBlkSize();
  GraphElem getBlkSize() const;
  GraphElem getBase(const int me) const;
  GraphElem getBound(const int me) const;
  int getOwner(const GraphElem v) const;
  PartRanges *parts;
  void setNumEdges(GraphElem numEdges); 
protected:
  DistGraph();
  DistGraph &operator = (const DistGraph &othis);
};

void loadDistGraphMPIIO(int me, int nprocs, int ranks_per_node, 
        DistGraph *&dg, std::string& fileName);

inline DistGraph::DistGraph()
  : totalNumVertices(0), totalNumEdges(0), localGraph(NULL), parts(NULL)
{
} // DistGraph

inline DistGraph::DistGraph(const GraphElem tnv, const GraphElem tne)
  : totalNumVertices(tnv), totalNumEdges(tne), localGraph(NULL), parts(NULL)
{
} // DistGraph

inline DistGraph::DistGraph(const DistGraph &othis)
  : totalNumVertices(othis.totalNumVertices), totalNumEdges(othis.totalNumEdges),
    localGraph(new Graph(*othis.localGraph)), parts(NULL)
{
  if (othis.parts)
    parts = new PartRanges(*othis.parts);
} // DistGraph

inline DistGraph::~DistGraph()
{
  if (localGraph)
    delete localGraph;
  if (parts)
    delete parts;
} // ~DistGraph

inline GraphElem DistGraph::getTotalNumVertices() const
{
  return totalNumVertices;
} // getTotalNumVertices

inline GraphElem DistGraph::getTotalNumEdges() const
{
  return totalNumEdges;
} // getTotalNumEdges

inline void DistGraph::createLocalGraph(const GraphElem lnv, const GraphElem lne,
					const PartRanges *oparts)
{
  assert(!localGraph);

  localGraph = new Graph(lnv, lne);
  if (oparts)
    parts = new PartRanges(*oparts);
} // createLocalGraph

inline Graph &DistGraph::getLocalGraph()
{
  assert(localGraph);

  return *localGraph;
} // getLocalGraph

inline const Graph &DistGraph::getLocalGraph() const
{
  assert(localGraph);

  return *localGraph;
} // getLocalGraph

inline GraphElem &DistGraph::getBlkSize()
{
  return blkSize;
} // getBlkSize

inline GraphElem DistGraph::getBlkSize() const
{
  return blkSize;
} // getBlkSize

inline GraphElem DistGraph::getBase(const int me) const
{
  if (parts)
    return parts->at(me);
  else
    return (static_cast<GraphElem>(me) * blkSize);
} // getBase

inline GraphElem DistGraph::getBound(const int me) const
{
  if (parts)
    return parts->at(me + 1);
  else {
    const GraphElem mbound = (static_cast<GraphElem>(me) + 1) * blkSize;
    const GraphElem rbound = std::min(mbound, totalNumVertices);

    return rbound;
  }
} // getBound

inline void DistGraph::setNumEdges(const GraphElem numEdges)
{
	this->totalNumEdges=numEdges;
}

inline int DistGraph::getOwner(const GraphElem v) const
{
  assert((v >= 0) && (v < totalNumVertices));

  if (parts) {  
    const PartRanges::const_iterator iter = std::upper_bound(parts->begin(), parts->end(), v);

    assert(iter != parts->end());
    return (iter - parts->begin() -1);
  }
  else
    return (v / blkSize);
} // getOwner

#endif // __DISTGRAPH_H
