// ***********************************************************************
//
//            Vite: A C++ library for distributed-memory graph clustering 
//                  using MPI+OpenMP
// 
//               Daniel Chavarria (daniel.chavarria@pnnl.gov)
//               Antonino Tumeo (antonino.tumeo@pnnl.gov)
//               Mahantesh Halappanavar (hala@pnnl.gov)
//               Pacific Northwest National Laboratory	
//
//               Hao Lu (luhowardmark@wsu.edu)
//               Sayan Ghosh (sayan.ghosh@wsu.edu)
//               Ananth Kalyanaraman (ananth@eecs.wsu.edu)
//               Washington State University
//
// ***********************************************************************
//
//       Copyright (2017) Battelle Memorial Institute
//                      All rights reserved.
//
// Redistribution and use in source and binary forms, with or without
// modification, are permitted provided that the following conditions
// are met:
//
// 1. Redistributions of source code must retain the above copyright
// notice, this list of conditions and the following disclaimer.
//
// 2. Redistributions in binary form must reproduce the above copyright
// notice, this list of conditions and the following disclaimer in the
// documentation and/or other materials provided with the distribution.
//
// 3. Neither the name of the copyright holder nor the names of its
// contributors may be used to endorse or promote products derived from
// this software without specific prior written permission.
//
// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
// "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
// FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
// COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
// INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
// LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
// LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
// ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
// POSSIBILITY OF SUCH DAMAGE.
//
// ************************************************************************

#ifndef __LOUVAIN_H
#define __LOUVAIN_H

#include <algorithm>
#include <fstream>
#include <functional>
#include <iostream>
#include <list>
#include <numeric>
#include <vector>
#include <unordered_map>
#include <unordered_set>
#include <map>

#include <mpi.h>
#include <omp.h>

#include "distgraph.hpp"
#include "coloring.hpp"

typedef std::vector<GraphElem> CommunityVector;
typedef std::vector<GraphWeight> GraphWeightVector;
typedef std::vector<GraphElem> GraphElemVector;

typedef std::unordered_map<GraphElem, GraphElem> VertexCommMap;

typedef std::unordered_set<GraphElem> RemoteVertexList;
typedef std::vector<RemoteVertexList> PartnerArray;

typedef std::vector<Comm> CommVector;
typedef std::map<GraphElem, Comm> CommMap;

typedef std::unordered_map<GraphElem, GraphElem> ClusterLocalMap;

const int SizeTag = 1;
const int VertexTag = 2;
const int CommunityTag = 3;
const int CommunitySizeTag = 4;
const int CommunityDataTag = 5;

struct CommInfo {
    GraphElem community;
    GraphElem size;
    GraphWeight degree;
};

// early termination cutoff for frozen 
// vertices is 90%
#define ET_CUTOFF 90
// early termination cutoff for 
// probabilistically freezing vertices 
// is 20%
#define P_CUTOFF 0.02

typedef std::vector<CommInfo> CommInfoVector;
extern std::ofstream ofs;
static MPI_Datatype commType;

double distLouvainMethod(const int me, const int nprocs, const DistGraph &dg,
        size_t &ssz, size_t &rsz, std::vector<GraphElem> &ssizes, 
        std::vector<GraphElem> &rsizes, std::vector<GraphElem> &svdata, 
        std::vector<GraphElem> &rvdata, CommunityVector &cvect, const double lower,
        const double thresh, int& iters);

double distLouvainMethod(const int me, const int nprocs, const DistGraph &dg,
        size_t &ssz, size_t &rsz, std::vector<GraphElem> &ssizes, 
        std::vector<GraphElem> &rsizes, std::vector<GraphElem> &svdata, 
        std::vector<GraphElem> &rvdata, CommunityVector &cvect, const double lower,
        const double thresh, int& iters, bool ETLocalOrRemote);

double distLouvainMethod(const int me, const int nprocs, const DistGraph &dg,
        size_t &ssz, size_t &rsz, std::vector<GraphElem> &ssizes, 
        std::vector<GraphElem> &rsizes, std::vector<GraphElem> &svdata, 
        std::vector<GraphElem> &rvdata, CommunityVector &cvect, const double lower,
        const double thresh, int& iters, double ETDelta, bool ETLocalOrRemote);

double distLouvainMethodWithColoring(const int me, const int nprocs, const DistGraph &dg,
        const long numColor, const ColorVector &vertexColor, size_t &ssz, size_t &rsz, 
        std::vector<GraphElem> &ssizes, std::vector<GraphElem> &rsizes, 
        std::vector<GraphElem> &svdata, std::vector<GraphElem> &rvdata,
        CommunityVector &cvect, const double lower, const double thresh, int& iters);

double distLouvainMethodWithColoring(const int me, const int nprocs, const DistGraph &dg,
        const long numColor, const ColorVector &vertexColor, size_t &ssz, size_t &rsz, 
        std::vector<GraphElem> &ssizes, std::vector<GraphElem> &rsizes, 
        std::vector<GraphElem> &svdata, std::vector<GraphElem> &rvdata,
        CommunityVector &cvect, const double lower, const double thresh, 
        int& iters, bool ETLocalOrRemote);

double distLouvainMethodWithColoring(const int me, const int nprocs, const DistGraph &dg,
        const long numColor, const ColorVector &vertexColor, 
        size_t &ssz, size_t &rsz, std::vector<GraphElem> &ssizes, 
        std::vector<GraphElem> &rsizes, std::vector<GraphElem> &svdata, 
        std::vector<GraphElem> &rvdata, CommunityVector &cvect, 
        const double lower, const double thresh, int& iters, double ETDelta, 
        bool ETLocalOrRemote);

double distLouvainMethodVertexOrder(const int me, const int nprocs, const DistGraph &dg,
        const long numColor, const ColorVector &vertexColor, size_t &ssz, size_t &rsz, 
        std::vector<GraphElem> &ssizes, std::vector<GraphElem> &rsizes, 
        std::vector<GraphElem> &svdata, std::vector<GraphElem> &rvdata,                        
        CommunityVector &cvect, const double lower, const double thresh, int& iters);

double distLouvainMethodVertexOrder(const int me, const int nprocs, const DistGraph &dg,
        const long numColor, const ColorVector &vertexColor, size_t &ssz, size_t &rsz, 
        std::vector<GraphElem> &ssizes, std::vector<GraphElem> &rsizes, 
        std::vector<GraphElem> &svdata, std::vector<GraphElem> &rvdata,                        
        CommunityVector &cvect, const double lower, const double thresh, int& iters, 
        bool ETLocalOrRemote);

double distLouvainMethodVertexOrder(const int me, const int nprocs, const DistGraph &dg,
        const long numColor, const ColorVector &vertexColor, size_t &ssz, size_t &rsz, 
        std::vector<GraphElem> &ssizes, std::vector<GraphElem> &rsizes, 
        std::vector<GraphElem> &svdata, std::vector<GraphElem> &rvdata,                        
        CommunityVector &cvect, const double lower, const double thresh, int& iters, 
        double ETDelta, bool ETLocalOrRemote);

static void distInitLouvain(const DistGraph &dg, CommunityVector &pastComm, 
        CommunityVector &currComm, GraphWeightVector &vDegree, 
        GraphWeightVector &clusterWeight, CommVector &localCinfo, 
        CommVector &localCupdate, double &constantForSecondTerm, 
        const int me);

static void distExecuteLouvainIteration(const GraphElem i, const DistGraph &dg,
        const CommunityVector &currComm, CommunityVector &targetComm,
        const GraphWeightVector &vDegree, CommVector &localCinfo, CommVector &localCupdate, 
        const VertexCommMap &remoteComm, const CommMap &remoteCinfo, CommMap &remoteCupdate,
        const double constantForSecondTerm, GraphWeightVector &clusterWeight, const int me);

static void distSumVertexDegree(const Graph &g, GraphWeightVector &vDegree, CommVector &localCinfo);

static double distCalcConstantForSecondTerm(const GraphWeightVector &vDegree);

static GraphElem distGetMaxIndex(const ClusterLocalMap &clmap, const GraphWeightVector &counter,
        const GraphWeight selfLoop, const CommVector &localCinfo, const CommMap &remoteCinfo,
        const GraphWeight vDegree, const GraphElem currSize, const GraphElem currDegree, 
        const GraphElem currComm, const GraphElem base, const GraphElem bound, const double constant);

static GraphWeight distBuildLocalMapCounter(const GraphElem e0, const GraphElem e1,
        ClusterLocalMap &clmap, GraphWeightVector &counter, const Graph &g, const CommunityVector &currComm,
        const VertexCommMap &remoteComm, const GraphElem vertex, const GraphElem base, const GraphElem bound);

static double distComputeModularity(const Graph &g, CommVector &localCinfo,
        const GraphWeightVector &clusterWeight, 
        const double constantForSecondTerm, const int me);

static void distUpdateLocalCinfo(CommVector &localCinfo, const CommVector &localCupdate);

static void distCleanCWandCU(const GraphElem nv, GraphWeightVector &clusterWeight,
        CommVector &localCupdate);

static void distInitComm(CommunityVector &pastComm, CommunityVector &currComm,
        const GraphElem base);

static void updateRemoteCommunities(const DistGraph &dg, CommVector &localCinfo,
        const CommMap &remoteCupdate,
        const int me, const int nprocs);

static void fillRemoteCommunities(const DistGraph &dg, const int me, 
        const int nprocs, const size_t &ssz, const size_t &rsz,
        const std::vector<GraphElem> &ssizes, const std::vector<GraphElem> &rsizes, 
        const std::vector<GraphElem> &svdata, const std::vector<GraphElem> &rvdata,
        const CommunityVector &currComm, const CommVector &localCinfo, 
        CommMap &remoteCinfo, VertexCommMap &remoteComm, CommMap &remoteCupdate);

static void exchangeVertexReqs(const DistGraph &dg, size_t &ssz, size_t &rsz,
        std::vector<GraphElem> &ssizes, std::vector<GraphElem> &rsizes, 
        std::vector<GraphElem> &svdata, std::vector<GraphElem> &rvdata,
        const int me, const int nprocs);

void createCommunityMPIType();
void destroyCommunityMPIType();

#endif
